//
// Raydialer NXT Interface Test V1.0 ( RobugC [wuerg] )
//
// Double frequency beacon detection:
//  "1" ->  697 + 1209  =>  CH1 + CH5
//  "5" ->  770 + 1336  =>  CH2 + CH6
//  "9" ->  862 + 1477  =>  CH3 + CH7
//
// (c)ASkr(FMMT666)
// www.askrprojects.net
//
//
// Balancer code by Laurens Valk (robotsquare.com)
//
//
// WARNING:
//   Just a quick example, no error checks.
//
//
// Does nothing else but aligning the brick with a beacon (front).
// The target beacon is changed every TIME_CHANGE_BEACON seconds.
//
//
//
// Today's Message: RobugC really sucks...
//
//
// timer resources:
//  - T4: balancer
//  - T3: beacon timer (stupid demo)
//

#pragma platform(NXT)



// Time (in s) after which the next beacon should be selected.
#define TIME_CHANGE_BEACON 10


// "include" the Raydialer-RX1 code
#include "asRaydRX1_01.c"


// for the remote control
#define REMOTE_CHANNEL 1


// for the balancer
const float your_wheel_diameter = 50;

const tSensors Gyro = S2;

int steering = 0;
int acceleration = 50;
int speed = 0;

float gn_dth_dt,gn_th,gn_y,gn_dy_dt,kp,ki,kd,mean_reading;

bool starting_balancing_task = true;

int calibrate_hitechnic();



//*************************************************************************************
//*** TaskBalancer
//***
//*** Author: Laurens Valk (robotsquare.com)
//*************************************************************************************
task TaskBalancer()
{
  // Set the time each loop cycle should last. You can set it up to 0.03 seconds or even higher, if you really
  // need to. If you add code to the control loop below (such as to read another sensor), make sure that
  // all code can run in under dt seconds. If it takes more time, set dt to a higher value.
  // Default is 0.010 seconds (10 miliseconds).
  float dt = 0.010;

  gn_dth_dt = 0.23;
  gn_th = 25.00;
  gn_y = 272.8;
  gn_dy_dt = 24.6;
  kp = 0.05;
  ki = 0.4;
  kd = 0.000504;


  //MOTOR SETUP
  nMotorPIDSpeedCtrl[motorB] = mtrNoReg;
  nMotorPIDSpeedCtrl[motorC] = mtrNoReg;
  nMotorEncoder[motorC] = 0;
  nMotorEncoder[motorB] = 0;


  SensorType[Gyro] = sensorRawValue;
  mean_reading = calibrate_hitechnic();


  //MATH CONSTANTS
  const float radius = your_wheel_diameter/1000;
  const float degtorad = PI/180;


  //SETUP VARIABLES FOR CALCULATIONS
  float u = 0,                    // Sensor Measurement (raw)
        th = 0,//Theta            // Angle of robot (degree)
        dth_dt = 0;//dTheta/dt    // Angular velocity of robot (degree/sec)
  float e = 0,//Error             // Sum of four states to be kept zero: th, dth_dt, y, dy_dt.
        de_dt = 0,//dError/dt     // Change of above error
        _edt = 0,//Integral Error // Accumulated error in time
        e_prev = 0;//Previous Error/ Error found in previous loop cycle
  float pid = 0;                  // SUM OF PID CALCULATION
  float y,//y                     // Measured Motor position (degrees)
        dy_dt,//dy/dt             // Measured motor velocity (degrees/sec)
	      v = 0,//velocity          // Desired motor velocity (degrees/sec)
	      y_ref = 0;//reference pos // Desired motor position (degrees)
  int motorpower = 0,             // Power ultimately applied to motors
      last_steering = 0,          // Steering value in previous cycle
      straight = 0,               // Average motor position for synchronizing
      d_pwr = 0;                  // Change in power required for synchronizing
  const int n_max = 7;            // Number of measurement used for floating motor speed average
  int n = 0,n_comp = 0,           // Intermediate variables needed to compute measured motor speed
  encoder[n_max];                 // Array containing last n_max motor positions
  starting_balancing_task = false;// We're done configuring. Main task now resumes.
  ClearTimer(T4);                 // This timer is used in the driver. Do not use it for other purposes!


  for(;;)
  {
    u = SensorRaw[Gyro];
    wait1Msec(2);
    u = u + SensorRaw[Gyro];

		//COMPUTE GYRO ANGULAR VELOCITY AND ESTIMATE ANGLE
  	dth_dt = u/2 - mean_reading;
  	mean_reading = mean_reading * 0.999 + (  0.001 * ( dth_dt + mean_reading )  );
  	th = th + ( dth_dt * dt );

    //ADJUST REFERENCE POSITION ON SPEED AND ACCELERATION

    if( v < speed * 10 )
      v = v + acceleration * 10 * dt;
    else
    {
      if(v > speed * 10)
        v = v - acceleration * 10 * dt;
    }
    y_ref = y_ref + v*dt;

  	//COMPUTE MOTOR ENCODER POSITION AND SPEED
  	n++;

  	if( n >= n_max )
  	  n = 0;

  	encoder[n] = nMotorEncoder[motorB] + nMotorEncoder[motorC] + y_ref;
  	n_comp = n+1;

  	if(n_comp == n_max)
  	  n_comp = 0;

  	y = encoder[n] * degtorad * radius;
  	dy_dt = ( encoder[n] - encoder[n_comp] ) / ( dt * ( n_max-1 ) ) * degtorad * radius;

  	//COMPUTE COMBINED ERROR AND PID VALUES
  	e = gn_th * th + gn_dth_dt * dth_dt + gn_y * y + gn_dy_dt * dy_dt;
  	de_dt = ( e - e_prev ) / dt;
  	_edt = _edt + e * dt;
  	e_prev = e;
  	pid = ( kp*e + ki*_edt + kd*de_dt) / radius;

  	//ADJUST MOTOR SPEED TO STEERING AND SYNCHING
    if(steering == 0)
    {
      if(last_steering != 0)
        straight = nMotorEncoder[motorC] - nMotorEncoder[motorB];

	    d_pwr = (nMotorEncoder[motorC] - nMotorEncoder[motorB] - straight) / (radius*10);
	  }
    else
      d_pwr = steering / ( radius * 10 );

    last_steering = steering;

  	//CONTROL MOTOR POWER AND STEERING
  	motorpower = pid;
    motor[motorB] = motorpower + d_pwr;
    motor[motorC] = motorpower - d_pwr;

    //ERROR CHECKING OR SHUTDOWN
    if( abs(th) > 60 || abs(motorpower) > 2000)
      StopAllTasks();

    //WAIT THEN REPEAT
  	while( time1[T4] < dt * 1000 )
  	  wait1Msec(1);

  	ClearTimer(T4);
  }// END for(;;)

}


//*************************************************************************************
//*** calibrate_hitechnic
//***
//*************************************************************************************
float calibrate_hitechnic()
{
  float mean = 0;
  int i;

  PlayTone( 220, 50 );
  eraseDisplay();
  nxtDisplayBigStringAt( 1, 50, "SENS CAL");

  for( i=0; i<30; i++)
  {
    mean += SensorRaw[Gyro];
    wait1Msec(100);
  }
  mean /= 30;

  eraseDisplay();
  nxtDisplayBigStringAt( 1, 60, "  MEAN");
  nxtDisplayBigStringAt( 1, 40, " %5.2f", mean);
  PlayTone( 440, 50 );
  wait10Msec(200);
  PlayTone( 880, 50 );

  return mean;
}





//*************************************************************************************
//*** screenUpdateRemote
//***
//*** Shows current values sent by the remote control.
//*************************************************************************************
void screenUpdateRemote( int ia, int ib, ubyte ic, string helper )
{
  int i;
  ubyte msk = 1;
  string stmp = "";
  char str[9];

  nxtDisplayBigStringAt( 1, 60, "DIR  PWR");
  nxtDisplayBigStringAt( 1, 40, "%3d  %3d", ia, ib);

  // yes, 3.5 ;-)
  nxtDisplayStringAt( 56 - ( 3.5*strlen(stmp) ), 10, helper );


  for( i=0; i<8; i++ )
  {
    if( ic & msk )
      str[i] = 'O';
    else
      str[i] = '-';
    msk <<= 1;
  }
  str[8] = 0;

  nxtDisplayBigStringAt( 1, 16, str );

}






//*************************************************************************************
//*** BeepBeep()
//***
//*************************************************************************************
void BeepBeep( int num )
{
  int i;

  for( i=0; i<num; i++ )
  {
    PlayTone( 440, 5 );
    wait10Msec(10);
   }
}




//*************************************************************************************
//*** main
//***
//*************************************************************************************
task main()
{
  ubyte btMsg[3] = {0,0,0};
  int ctrlSteer = 0;
  int ctrlPower = 0;
  ubyte ctrlDigi = 0;

  int i;

  // increase volume (for BeepBeep() only...)
  nVolume = 4;

  // init Raydialer-RX
  RaydInit();

  // start balancing
  StartTask(TaskBalancer);

  // wait until we are ready
  while( starting_balancing_task )
  {;}



  // ------------------------------------------------------------------------------------
  // STANDARD BEHAVIOUR IF ORANGE BUTTON IS NOT PRESSED:
  //   - change target beacon every 4 seconds
  //   - steer until beacon is straight ahead (very stupid "algorithm")


  if( nNxtButtonPressed != 3 )
  {
    eraseDisplay();
    nxtDisplayBigStringAt( 1, 60, " REMOTE ");
    nxtDisplayBigStringAt( 1, 40, "  OFF!  ");

    wait10Msec(100);
    eraseDisplay();

    i = 0;            // current beacon
    ClearTimer(T3);   // timer to determine beacon target change

    for(;;)
    {
      BeacFind( BEAC_THRES );

      if( time100[T3] > TIME_CHANGE_BEACON * 10 )
      {
        i++;
        if( i >= BEAC_NUM )
          i = 0;
        BeepBeep( i + 1 );
        ClearTimer(T3);
      }

      if( gBeacAng[i] != 666 )
      {
/*
        It would be wise to implement a PID alg for steering:

		    e = w - x;
		    esum = esum + e;
		    y = Kp*e + Ki*Ta*esum + Kd/Ta*(e  ealt);
		    ealt = e;
*/

        steering = gBeacAng[i] / 3;
      }
      else
        steering = 0;

    }
  }


  // ------------------------------------------------------------------------------------
  // ALTERNATIVE BEHAVIOUR IF ORANGE BUTTON IS PRESSED:
  //   - Receive movement commands from other brick.
  //     For transmitter source code, see webpage
  //     (or maybe ZIP file).
  // Bricks need to be linked and transmitter program must
  // be started before this code is executed.
  for(;;)
  {
    // receive 3 bytes in this order:
    //  1) steer
    //  2) power
    //  3) buttons (digital channel, one bit per channel)
    cCmdMessageRead(btMsg, 3, REMOTE_CHANNEL);
    ctrlSteer = ( btMsg[0] - 128 ) / 3;
    ctrlPower = btMsg[1] - 128;
    ctrlDigi  = btMsg[2];

    screenUpdateRemote( ctrlSteer, ctrlPower, ctrlDigi, "" );

    wait1Msec(100);

    steering = ctrlSteer;
    speed = ctrlPower;
  }

}
