//
// NOTICE:
//   DO NOT START/COMPIILE THIS DIRECTLY.
//   RobugC does not support more than one file in a "project".
//   Include this one ("asRaydRX1_xy.c") in your main module via:
//
//     #include "asRaydRX1_xy.c"
//
//
//
// asRaydRX1 V0.1
//
// Raydialer-RX1 interface demonstration for RobugC.
// A DTMF, dual frequency local positioning experiment.
//
// (c)2012 ASkr(FMMT666)
// www.askrprojects.net
//
//
//
//
// NOTE #0:
//   Watch out for Raydialer-RX2 (and TX2), which support more
//   unique channels and much more...
//
// NOTE #1:
//   Due to RobugC and its limitations, only nine tones are supported.
//   Beacons "A", "B", "C" and "D" can not be used.
//
// NOTE #2:
//   This demo uses ONLY three beacons: "1", "5" and "9".
//
// NOTE #3:
//   For restrictions and usage, see webpage.
//
// NOTE #4:
//   This source does not contain any sophisticated analyses.
//   It's only a demo (and Raydialer only an experiment ;-)
//
//
//
//      Complete DTMF Frequency Table
//    ==================================
//
//           1209Hz 1336Hz 1477Hz 1633Hz
//     697Hz    1      2      3      A
//     770Hz    4      5      6      B
//     852Hz    7      8      9      C
//     941Hz    *      0      #      D
//
//
//      RAYDIALER TONE ORDER
//    =======================
//     see (*2*) in code below
//
//     NUM  FREQUENCY[Hz]
//      0       697
//      1       770
//      2       852
//      3       941
//      4      1209
//      5      1336
//      6      1477
//      7      1633
//
//
//       RAYDIALER BEACON ORDER
//    =============================
//
//     BEACON# TONE FREQUENCIES[Hz]
//        0      1  697 + 1209
//        1      2  697 + 1336
//        2      3  697 + 1477
//        3      4  770 + 1209
//        4      5  770 + 1336
//        5      6  770 + 1477
//        6      7  852 + 1209
//        7      8  852 + 1336
//        8      9  852 + 1477
//        9      0  941 + 1336
//       10      *  941 + 1209
//       11      #  941 + 1477
//       12      A  697 + 1633
//       13      B  770 + 1633
//       14      C  852 + 1633
//       15      D  941 + 1633
//
//
//
//
// Inside your code, you'll only need:
//
//   RaydInit();                  // to init Raydialer-RX1
//
//   BeacFind( BEAC_THRES );      // to start a measurement (fills gBeacAng[] array)
//
//   whatever = gBeacAng[<bnum>]  // analyze the angle output of a beacon
//                                //  0  = beacon straight ahead
//                                // <0  = beacon to left
//                                // >0  = beacon to right
//                                // 666 = no beacon
//
// Maximum beacon angle is defined by BEAC_ANGLE2.
// You may wish to tune this (and BEAC_ANGLE1).
//


const tSensors sRAYD = S3;


#define BEAC_NUM          3               // number of beacons
#define BEAC_THRES        100             // default beacon threshold
#define BEAC_ANGLE1       20              // absolute angle to 1st receiver
#define BEAC_ANGLE2       60              // absolute angle to 2nd receiver
#define BEAC_OFF          666             // no-beacon-deteced-marker in "angle-array" gBeacAng[]

#define RAYD_TONES        7               // number of tones

#define RAYD_I2C_ABYTE 0x90               // base address byte (not I2C address)
#define RAYD_I2C_RLEN  (2*RAYD_TONES+1)   // 1 status byte + 7x2 channel bytes

#define RAYD_CMD_STOP  27                 // ESC -> stop/abort
#define RAYD_CMD_DTMF  'D'                // 'D' -> DTMF single
#define RAYD_CMD_MAG   'o'                // 'o' -> magnitude output


// index of tones in a (pseudo) array of 8 tone levels
// see (*1*) above
const ubyte cBeacChans[BEAC_NUM][2]={ {0,4}, {1,5}, {2,6} };

// angular displacements of receivers
const int cBeacAngles[4]={ -BEAC_ANGLE2, -BEAC_ANGLE1, BEAC_ANGLE1, BEAC_ANGLE2 };


// Using RobugC? Fuct...
// I2C buffer: 32 bytes planned, 17 needed, but more than 16 can not be
// used (Did I already mention what I think of RobotC?).
// Hence, dialtones "A..D" can not be used in here...
// Still no pointers...
ubyte gI2CBufO[20]; // add some headroom
ubyte gI2CBufI[20];


// "ROBOTC only supports 'unsigned' qualifier on 'byte' / 'char'
//  sized variables. Ignored for other type."
// Ridiculous, like the rest ;-) And due to some
// other flaws we really needed >longs< here.
// Getting tired of "coding" RobugC; no defines any more...
long gBeacLev[BEAC_NUM][4];   // 3 beacons and 4 receivers

int  gBeacAng[BEAC_NUM];      // calculated angle of beacon -90..90 (0 = front; 666 = none)

long gToneLev[4][RAYD_TONES]; // 4 receivers and 7 tones




//***************************************************************************
//*** I2CWrite()
//***
//*** I2C write utility function.
//***************************************************************************
int I2CWrite()
{
  while( nI2CStatus[sRAYD] == STAT_COMM_PENDING )
  {;}
  sendI2CMsg( sRAYD, gI2CBufO[0], 0 );

  // TODO: I2C return
  return 0;
}



//***************************************************************************
//*** I2CRead()
//***
//*** I2C read utility function.
//***************************************************************************
int I2CRead( int ans )
{
  while( nI2CStatus[sRAYD] == STAT_COMM_PENDING )
  {;}
  sendI2CMsg( sRAYD, gI2CBufO[0], ans );

  while (nI2CStatus[sRAYD] == STAT_COMM_PENDING)
  {;}
  readI2CReply( sRAYD, gI2CBufI[0], ans );


  // TODO: I2C return
  return 0;
}


//***************************************************************************
//*** RaySendCmd()
//***
//*** Sends command <cmd> to Raydialer.
//*** For a complete list of commands, see webpage.
//***************************************************************************
int RaydSendCmd( int cmd )
{
  gI2CBufO[0] = 0x02;
  gI2CBufO[1] = RAYD_I2C_ABYTE;
  gI2CBufO[2] = cmd & 0xff;
  I2CWrite();

  // TODO: I2C return
  return 0;
}



//***************************************************************************
//*** RaydGetChannel()
//***
//*** Fetches data from Raydialer RX, calcs tone levels and stores
//*** them inside the global array "gToneLev".
//***************************************************************************
int RaydGetChannel( int ch )
{
  int i;

  gI2CBufO[0] = 0x01;
  gI2CBufO[1] = RAYD_I2C_ABYTE | 0x01 | ( (ch & 0x03) << 1 );

  I2CRead( RAYD_I2C_RLEN );

  for( i=0; i<RAYD_TONES; i++ )
    gToneLev[ch][i] = ( (long)gI2CBufI[(2*i)+1] << 8 ) + gI2CBufI[(2*i)+2];

  // TODO: I2C return
  return 0;
}



//***************************************************************************
//*** RaydInit()
//***
//*** Initializes Raydialer
//***************************************************************************
int RaydInit()
{
  SensorType[sRAYD] = sensorI2CCustomFast;

  wait10Msec(10);
  RaydSendCmd( RAYD_CMD_STOP );
  wait10Msec(50);
  RaydSendCmd( RAYD_CMD_MAG );
  wait10Msec(10);

  // TODO: RaydInit return
  return 0;
}



//***************************************************************************
//*** BeacCalcLev()
//***
//*** Finds beacon level per receiver (channel) in gToneLev[<receiver>][<tones>]
//*** and fills the global array gBeacLev[<beacon>][<channel>].
//***************************************************************************
int BeacCalcLev( int threshold, int ch )
{
  int i;
  long li1, li2;

  for( i=0; i<BEAC_NUM; i++ )
  {
    li1 = gToneLev[ch][ cBeacChans[i][0] ];
    li2 = gToneLev[ch][ cBeacChans[i][1] ];

    // TODO: beacon 2-tone match check
    if( ( li1 > threshold ) && ( li2 > threshold ) )
    {
      // preliminary: average (instead of power) should be sufficient, right now
      gBeacLev[i][ch] = ( li1 + li2 ) / 2;
    }
    else
      gBeacLev[i][ch] = 0;
  }


  // TODO: BeacCalcLev return
  return 0;
}




//***************************************************************************
//*** BeacCalcAng()
//***
//*** Uses gBeacLev[<beacon>][<channel>]. to calculate the angular
//*** displacement of each beacon and stores the results in gBeacAng[].
//*** If no beacon has been detected, it is marked with value "BEAC_OFF".
//***************************************************************************
int BeacCalcAng( )
{
  int i, j;
  long li1, li2;

  for( i=0; i<BEAC_NUM; i++ )
  {
    li1 = 0;
    li2 = 0;

	  for( j=0; j<4; j++ )
	  {
	    li2 += gBeacLev[i][j];
	    li1 += ( cBeacAngles[j] * gBeacLev[i][j] );
	  }

	  if ( (li1 /= li2) == 0 )
	    gBeacAng[i] = BEAC_OFF;
	  else
	    gBeacAng[i] = li1;

  }// END for each beacon


  // TODO: BeacCalcLev return
  return 0;
}



//***************************************************************************
//*** BeacFind()
//***
//***
//***************************************************************************
int BeacFind( int thres )
{
  int i;

  // start scan (~200ms)
  RaydSendCmd( RAYD_CMD_DTMF );
  wait10Msec(25);

  for( i=0; i<4; i++ )
  {
    RaydGetChannel(i);
    BeacCalcLev( thres, i );

    wait10Msec(1);
  }

  BeacCalcAng();

  // TODO: FindBeacon return
  return 0;
}
